/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  DSP - digital signal processor interface (0xcc005000)
 *  AI  - audio interface (0xcc006c00)    
 *         
 */


#include "hw_ai_dsp.h"


__u8 rdsp[RDSP_SIZE] = {0};
__u8 rai[RAI_SIZE] = {0};


__u16 dsp_r16_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, ".dsp: read  [%.4x] (%.4x)", addr & 0xffff, RDSP16 (addr));
	return RDSP16 (addr);
}


void dsp_w16_direct (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG, ".dsp: write [%.4x] (%.4x) = %.4x", addr & 0xffff, RDSP16 (addr), data);
	RDSP16 (addr) = data;
}


__u32 dsp_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, ".dsp: read  [%.4x] (%.8x)", addr & 0xffff, RDSP32 (addr));
	return RDSP32 (addr);
}


void dsp_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, ".dsp: write [%.4x] (%.8x) = %.8x", addr & 0xffff, RDSP32 (addr), data);
	RDSP32 (addr) = data;
}


__u32 ai_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, "..ai: read  [%.4x] (%.8x)", addr & 0xffff, RAI32 (addr));
	return RAI32 (addr);
}


void ai_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..ai: write [%.4x] (%.8x) = %.8x", addr & 0xffff, RAI32 (addr), data);
	RAI32 (addr) = data;
}


void ai_w32_cr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..ai: AICR | PSTAT %d | FREQ %d kHz | AIINT %d/%d | AIINTVLD %d | SCRESET %d",
				 data & AICR_PSTAT, ((data >> 1) & 1) ? 48 : 32, (data & AICR_AIINTMSK)>0, (data & AICR_AIINT)>0, (data & AICR_AIINTVLD)>0, (data & AICR_SCRESET)>0);

	audio_set_freq ((data & AICR_32KHZ) ? FREQ_32KHZ : FREQ_48KHZ);

	if (data & AICR_SCRESET)
	{
		AISCNT = 0;
		data &= ~AICR_SCRESET;
	}

	RAI32 (addr) = data;
}


// sample counter
__u32 ai_r32_aiscnt (__u32 addr)
{
	DEBUG (EVENT_LOG, "..ai: AISCNT read %.8x", RAI32 (addr));

	// fake it
	return RAI32 (addr)++;
}


void ai_w32_aivr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..ai: VOLUME %d / %d", data & 0x00ff, data & 0xff00);
	RAI32 (addr) = data;
}


void dsp_w16_csr (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG, ".dsp: CSR = %.4x %s %s", data,
				 (data & DSP_CSR_RES) ? "reset" : "", (data & DSP_CSR_HALT) ? "halt" : "");

	if (data & DSP_CSR_PIINT)
		pi_interrupt (INTERRUPT_DSP);

	if (data & DSP_CSR_DSPINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_DSPINT;

	if (data & DSP_CSR_ARINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_ARINT;

	if (data & DSP_CSR_AIDINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_AIDINT;

	RDSP16 (DSP_CSR) |= data & (DSP_CSR_DSPINTMSK | DSP_CSR_ARINTMSK | DSP_CSR_AIDINTMSK);
}


__u16 dsp_r16_mailbox (__u32 addr)
{
	return (RDSP16 (addr) ^= 0x8000);
}


void dsp_w16_mailbox (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG, ".dsp: MAIL %.4x", data);
}


void dsp_w16_dmacr (__u32 addr, __u16 data)
{
	RDSP16 (addr) = data;

	if (data & DMACR_PLAY)
	{
		DEBUG (EVENT_LOG, ".dsp: DMACR start %.8x length %d",
					 (RDSP16 (DSP_DMA_START_H) << 16) | RDSP16 (DSP_DMA_START_L) | 0x80000000,
					 (RDSP16 (DSP_DMACR) & 0x7fff) << 5);

		audio_play (MEM_ADDRESS ((RDSP16 (DSP_DMA_START_H) << 16) | RDSP16 (DSP_DMA_START_L)),
								(data & 0x7fff) << 5);
	}
	else
	{
		audio_stop ();
		DEBUG (EVENT_LOG, ".dsp: DMACR stop, length %d", (data & 0x7fff) << 5);
	}
}


void dsp_w32_arcnt (__u32 addr, __u32 data)
{
	RDSP32 (addr) = data;

	if (data & ~ARCNT_READ)
	{
		// transfer to aram
		DEBUG (EVENT_LOG, ".dsp: write %d MEM[%.8x] -> ARAM[%.8x]",
					 data & ~ARCNT_READ, RDSP32 (ARDMA_MMADDR) | 0x80000000, RDSP32 (ARDMA_ARADDR));

		RDSP16 (DSP_CSR) |= DSP_CSR_ARINT;
		if (RDSP16 (DSP_CSR) & DSP_CSR_ARINTMSK)
			pi_interrupt (INTERRUPT_DSP);
	}
	else
		DEBUG (EVENT_LOG, ".dsp: read %d MEM[%.8x] <- ARAM[%.8x]",
					 data & ~ARCNT_READ, RDSP32 (ARDMA_MMADDR) | 0x80000000, RDSP32 (ARDMA_ARADDR));
}


void dsp_finished (void)
{
	DEBUG (EVENT_LOG, ".dsp: finished");

	RDSP16 (DSP_CSR) |= DSP_CSR_AIDINT;
	if (RDSP16 (DSP_CSR) & DSP_CSR_AIDINTMSK)
		pi_interrupt (INTERRUPT_DSP);

	RDSP16 (DSP_DMACR) &= ~DMACR_PLAY;
}


void dsp_init (void)
{
	memset (rdsp, 0, sizeof (rdsp));
	memset (rai, 0, sizeof (rai));

	RAI32 (AI_AICR) |= AICR_32KHZ;

	// control status register
	mem_hwr_hook (16, DSP_CSR, dsp_r16_direct);
	mem_hww_hook (16, DSP_CSR, dsp_w16_csr);

	// dsp mailbox
	mem_hwr_hook (16, 0x5000, dsp_r16_mailbox);
	mem_hww_hook (16, 0x5000, dsp_w16_mailbox);
	mem_hwr_hook (16, 0x5002, dsp_r16_direct);
	mem_hww_hook (16, 0x5002, dsp_w16_mailbox);

	// cpu mailbox
	mem_hwr_hook (16, 0x5004, dsp_r16_mailbox);
	mem_hww_hook (16, 0x5004, dsp_w16_mailbox);
	mem_hwr_hook (16, 0x5006, dsp_r16_direct);
	mem_hww_hook (16, 0x5006, dsp_w16_mailbox);

	// ???
	RDSP16 (0x5016) = 1;
	mem_hwr_hook (16, 0x5016, dsp_r16_direct);
	
	mem_hwr_hook (32, ARDMA_MMADDR, dsp_r32_direct);
	mem_hww_hook (32, ARDMA_MMADDR, dsp_w32_direct);
	mem_hwr_hook (32, ARDMA_ARADDR, dsp_r32_direct);
	mem_hww_hook (32, ARDMA_ARADDR, dsp_w32_direct);

	// ARDMA CNT
	mem_hww_hook (32, 0x5028, dsp_w32_arcnt);

	// DMA start / end address
	mem_hww_hook (16, 0x5030, dsp_w16_direct);
	mem_hwr_hook (16, 0x5030, dsp_r16_direct);
	mem_hww_hook (16, 0x5032, dsp_w16_direct);
	mem_hwr_hook (16, 0x5032, dsp_r16_direct);
	
	// DMA control / length
	mem_hww_hook (16, 0x5036, dsp_w16_dmacr);
	mem_hwr_hook (16, 0x5036, dsp_r16_direct);
	
	// DMA bytes left
	mem_hwr_hook (16, 0x503a,	dsp_r16_direct);

	// ai control register
	mem_hwr_hook (32, AI_AICR, ai_r32_direct);
	mem_hww_hook (32, AI_AICR, ai_w32_cr);
	
	// ai sample counter
	mem_hwr_hook (32, AI_AISCNT, ai_r32_aiscnt);

	// volume
	mem_hww_hook (32, AI_VR, ai_w32_aivr);
	mem_hwr_hook (32, AI_VR, ai_r32_direct);

	// ai interrupt timing
	mem_hww_hook (32, AI_IT, ai_w32_direct);
}
