#ifndef __HW_DSP_H
#define __HW_DSP_H 1


#include "general.h"
#include "mem.h"

#define RDSP_SIZE					0x200
#define RDSP_MASK					0x1ff

#define RDSP16(X)					(*((__u16 *) &rdsp[RDSP_SIZE - ((X) & RDSP_MASK) - 2]))
#define RDSP32(X)					(*((__u32 *) &rdsp[RDSP_SIZE - ((X) & RDSP_MASK) - 4]))

#define RAI_SIZE					0x200
#define RAI_MASK					0x1ff

#define RAI32(X)					(*((__u32 *) &rai[RAI_SIZE - ((X) & RDSP_MASK) - 4]))


extern __u8 rdsp[RDSP_SIZE];
extern __u8 rai[RAI_SIZE];


#define DSP_CSR							0x500a
#define DSP_CSR_RES					(1 << 0)
#define DSP_CSR_PIINT				(1 << 1)
#define DSP_CSR_HALT				(1 << 2)
#define DSP_CSR_AIDINT			(1 << 3)
#define DSP_CSR_AIDINTMSK		(1 << 4)
#define DSP_CSR_ARINT				(1 << 5)
#define DSP_CSR_ARINTMSK		(1 << 6)
#define DSP_CSR_DSPINT			(1 << 7)
#define DSP_CSR_DSPINTMSK		(1 << 8)
#define DSP_CSR_DMAINT			(1 << 9)

#define AI_AICR							0x6c00
#define AICR								(RAI32 (AI_AICR))
#define AICR_SCRESET				(1 << 5)
#define AICR_AIINTVLD				(1 << 4)
#define AICR_AIINT					(1 << 3)
#define AICR_AIINTMSK				(1 << 2)
#define AICR_PSTAT					(1 << 0)

//
#define AICR_32KHZ					(1 << 6)

// volume
#define AI_VR								0x6c04

#define AI_AISCNT						0x6c08
#define AISCNT							(RAI32 (AI_AISCNT))
#define AI_IT								0x6c0c

#define ARCNT_READ					0x80000000

#define ARDMA_MMADDR				0x5020
#define ARDMA_ARADDR				0x5024

#define DSP_DMA_START_H			0x5030
#define DSP_DMA_START_L			0x5032
#define DSP_DMACR						0x5036
#define DSP_DMA_BLOCKS_LEFT	0x503a
#define DMACR_PLAY					(0x8000)

// len >> 5
#define STREAM_BLOCKS_LEFT		(RDSP16 (DSP_DMA_BLOCKS_LEFT))
#define STREAM_FINISHED				(RDSP16 (DSP_DMACR) &= ~DMACR_PLAY)


void dsp_init (void);
void dsp_finished (void);


#endif // __HW_DSP_H
