/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  EXI - external interface (0xcc006800)
 *      
 *         
 */



#include "hw_exi.h"


__u8 rexi[REXI_SIZE] = {0};



__u32 exi_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, ".exi: read  [%.4x] (%.8x)", addr & 0xffff, REXI32 (addr));
	return REXI32 (addr);
}


void exi_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, ".exi: write [%.4x] (%.8x) = %.8x", addr & 0xffff, REXI32 (addr), data);
	REXI32 (addr) = data;
}


// control status register
void exi_w32_csr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, ".exi: CSR%d | CLK %d%d%d | DEV %d%d%d | EXT %d | INT %d%d%d%d%d%d",
				 (addr & 0xff) / 0x14,	(data & 0x40)>0, (data & 0x20)>0, (data & 0x10)>0,
				 (data >> 9) & 1, (data >> 8) & 1, (data >> 7) & 1, (data >> 12) & 1,
				 (data & EXI_CSR_EXTINT)>0, (data & EXI_CSR_EXTINTMSK)>0, (data & EXI_CSR_TCINT)>0, (data & EXI_CSR_TCINTMSK)>0, (data & EXI_CSR_EXTINT)>0, (data & EXI_CSR_EXTINTMSK)>0);

	REXI32 (addr) = data;
}


void exi_w32_cr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, ".exi: CR%d | TLEN %d%d | RW %d%d | DMA %d | START %d", (addr & 0xff) / 0x14,
				 (data >> 5) & 1, (data >> 4) & 1, (data >> 3) & 1, (data >> 2) & 1, (data >> 1) & 1, data & 1);

	if (data & EXI_CR_TSTART)
	{
		// csr, transfer complete interrupt
		REXI32 (addr - 0x0c) |= EXI_CSR_TCINT;
		if (REXI32 (addr - 0x0c) & EXI_CSR_TCINTMSK)
			pi_interrupt (INTERRUPT_EXI);
	}
}


void exi_init (void)
{
	memset (rexi, 0, sizeof (rexi));

	mem_hwr_hook (32, EXI_0CSR, exi_r32_direct);
	mem_hww_hook (32, EXI_0CSR, exi_w32_csr);
	mem_hwr_hook (32, EXI_0MAR, exi_r32_direct);
	mem_hww_hook (32, EXI_0MAR, exi_w32_direct);
	mem_hwr_hook (32, EXI_0LENGTH, exi_r32_direct);
	mem_hww_hook (32, EXI_0LENGTH, exi_w32_direct);
	mem_hww_hook (32, EXI_0CR, exi_w32_cr);
	mem_hwr_hook (32, EXI_0CR, exi_r32_direct);
	mem_hwr_hook (32, EXI_0DATA, exi_r32_direct);
	mem_hww_hook (32, EXI_0DATA, exi_w32_direct);

	mem_hwr_hook (32, EXI_1CSR, exi_r32_direct);
	mem_hww_hook (32, EXI_1CSR, exi_w32_direct);
	mem_hww_hook (32, EXI_1CR, exi_w32_cr);
	mem_hwr_hook (32, EXI_1CR, exi_r32_direct);
	mem_hwr_hook (32, EXI_1DATA, exi_r32_direct);
	mem_hww_hook (32, EXI_1DATA, exi_w32_direct);

	mem_hwr_hook (32, EXI_2CSR, exi_r32_direct);
	mem_hww_hook (32, EXI_2CSR, exi_w32_direct);
	mem_hww_hook (32, EXI_2CR, exi_w32_cr);
	mem_hwr_hook (32, EXI_2CR, exi_r32_direct);
	mem_hwr_hook (32, EXI_2DATA, exi_r32_direct);
	mem_hww_hook (32, EXI_2DATA, exi_w32_direct);
}
