/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  GX FIFO - graphic display list (0xcc006000)
 *  this is quick and ugly, will be cleaned up later
 *         
 */

#include "hw_gx.h"

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#ifndef LOG_GX
# undef DEBUG
# define DEBUG(X,format,...) {}
#endif

int use_colors = TRUE;
int use_textures = TRUE;

// Vertex Descriptor (VCD)
#define VCD_NONE					0
#define VCD_DIRECT				1
#define VCD_INDEXED8			2
#define VCD_INDEXED16			3

int vcd_pos, vcd_normal, vcd_col0, vcd_col1, vcd_tex0;

#define VAT_FLOAT					0
#define VAT_U32						1
#define VAT_U16						2
#define VAT_S16						3

#define VAT_TEX_U8				0
#define VAT_TEX_U16				2
#define VAT_TEX_S16				3

unsigned int glfilter[] = { GL_NEAREST, GL_LINEAR };
#define TEX_MODE_MIN_FILTER(X)					(glfilter[(X >> 5) & 1])
#define TEX_MODE_MAG_FILTER(X)					(glfilter[(X >> 4) & 1])

unsigned int glwrap[] = { GL_CLAMP, GL_CLAMP_TO_EDGE, GL_REPEAT };
#define TEX_MODE_WRAP_S(X)							(glwrap[X & 3])
#define TEX_MODE_WRAP_T(X)							(glwrap[(X >> 2) & 3])


// Vertex Attribute (VAT)
int vat_pos_format, vat_normal_format, vat_col0_format, vat_col1_format, vat_tex0_format;
int vat_pos_count, vat_normal_count, vat_col0_count, vat_col1_count, vat_tex0_count;

unsigned int tex_width, tex_height, tex_format, tex_mode;
unsigned int textures[4096];
__u32 mtextures[4096];
int ntextures = 0;


// incorrect
#define CIA4						10
//#define CI8							20
#define CIA8						30
//#define RGB4A3					40
#define YCbYCr					60
//#define RGB565					70
// just guessing
#define CI4							0
#define CI8							1
#define RGB4A3					6
#define RGB565					4
// correct
#define RGB5A1					5


__u8 palette[255] = {0};

void put_tile (unsigned char *dest, int x, int y, unsigned char *src, int w,
							 int h, int tx, int ty, int format)
{
	int i, j;
	__u16 *p = (__u16 *) src;
	__u16 d;


	dest += y*w*4 + x*4;
	for (j = 0; j < ty; j++)
	{
		for (i = 0; i < tx; i++)
		{
			d = BSWAP16 (*p);

			// most of it is just greyscale for now
			switch (format)
			{
				case CI4:
					// CI4 iiii
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = (d & 0xf000) >> 8;
					dest[4*i + 3] = 0xff;

					i++;
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = (d & 0x0f00) >> 4;
					dest[4*i + 3] = 0xff;

					i++;
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = (d & 0x00f0) >> 0;
					dest[4*i + 3] = 0xff;

					i++;
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = (d & 0x000f) << 4;
					dest[4*i + 3] = 0xff;
					break;
			
				// CI8 iiii iiii
				case CI8:
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = palette [(d & 0xff00) >> 6];
					dest[4*i + 3] = 0xff;

					i++;
					dest[4*i + 0] =
					dest[4*i + 1] =
					dest[4*i + 2] = palette [(d & 0x00ff) << 2];
					dest[4*i + 3] = 0xff;
					break;
			
				case YCbYCr:
					// YCbYCr Y1 Cb Y2 Cr
					dest[4*i + 0] = 
					dest[4*i + 1] = 
					dest[4*i + 2] = (d & 0xff00) >> 8;
					dest[4*i + 3] = 0xff;
					break;
			
				case RGB4A3:
					// RGB4A3  .aaa rrrr gggg bbbb
					dest[4*i + 0] = (d & 0x0f00) >> 4;
					dest[4*i + 1] = (d & 0x00f0) >> 0;
					dest[4*i + 2] = (d & 0x000f) << 4;
					dest[4*i + 3] = (d & 0x7000) >> 7;
					break;

				case RGB5A1:
					// RGB5A1  arrr rrgg gggb bbbb
					dest[4*i + 0] = (d & 0x7c00) >> 7;
					dest[4*i + 1] = (d & 0x03e0) >> 2;
					dest[4*i + 2] = (d & 0x001f) << 3;
					dest[4*i + 3] = (d & 0x8000) >> 8;
					break;

				case RGB565:
					// RGB565  rrrr rggg gggb bbbb
					dest[4*i + 0] = (d & 0xf800) >> 8;
					dest[4*i + 1] = (d & 0x07e0) >> 3;
					dest[4*i + 2] = (d & 0x001f) << 3;
					dest[4*i + 3] = 0xff;
					break;
			}

			p++;
		}
		dest += w*4;
	}
}


char *gx_convert_texture (char *addr, int w, int h, int format)
{
	static unsigned char buff[1024*1024*4];
	int corner_x = 0, corner_y = 0;
	int tx, ty, bpp;


	// trying to get some visual
	{
		int i;
		palette[0] = 0x00;
		for (i = 1; i < 255; i++)
			palette[i] = palette[i-1] + 0x11;
	}
	
	switch (format)
	{
		// palette index, 4 bits per index, 8x4 tiles
		case CI4:
			tx = 8;
			ty = 4;
			bpp = 4;
			break;

		// palette index, 8 bits per index, 8x4 tiles
		case CI8:
			tx = 8;
			ty = 4;
			bpp = 8;
			break;
		
		// 16 bits per pixel, 4x4 tiles
		case RGB4A3:
		case RGB5A1:
		case RGB565:
			tx = 4;
			ty = 4;
			bpp = 16;
			break;
		
		default:
			return NULL;
	}
	
	corner_x = 0;
	corner_y = 0;
	while (corner_y < h)
	{
		put_tile (buff, corner_x, corner_y, addr, w, h, tx, ty, format);

		addr += tx * ty * bpp/8;
		corner_x += tx;
		if (corner_x == w)
		{
			corner_x = 0;
			corner_y += ty;
		}
	}

	return buff;
}


int push_vertices (int n, int addr)
{
	int i, k = 0, l = 0;


	for (i = 0; i < n; i++)
	{
		// vertex position setup
		// glVertex must the last one called
		switch (vat_pos_format)
		{
			case VAT_FLOAT:
				k += 12;
				break;

			case VAT_U16:
			case VAT_S16:
				if (vat_pos_count == 0)
					k += 4;
				else
					k += 6;
				break;
		}

		if (vcd_col0)
		{
			switch (vat_col0_format)
			{
				case VAT_U32:
					DEBUG (EVENT_LOG, "....  color %8.8x", MEMR32 (addr + k));
					if (use_colors)
						glColor4ub (MEM (addr + k + 0), MEM (addr + k + 1), MEM (addr + k + 2), MEM (addr + k + 3));
					k += 4;
					break;

				default:
					DEBUG (EVENT_EFATAL, "!!!! unemulated color VAT %d", vat_col0_format);
			}
		}

		if (vcd_tex0)
		{
			switch (vat_tex0_format)
			{
				case VAT_TEX_U8:
					DEBUG (EVENT_LOG, "....  tex (%d, %d)", MEM (addr + k + 0), MEM (addr + k + 1));
					glTexCoord2s (MEM (addr + k + 0), MEM (addr + k + 1));
					k += 2;
					break;

				case VAT_TEX_U16:
					DEBUG (EVENT_LOG, "....  tex (%d, %d)", MEMR16 (addr + k + 0), MEMR16 (addr + k + 2));
					glTexCoord2s (MEMR16 (addr + k + 0), MEMR16 (addr + k + 2));
					k += 4;
					break;

				case VAT_TEX_S16:
					DEBUG (EVENT_LOG, "....  tex (%d, %d)", MEMSR16 (addr + k + 0), MEMSR16 (addr + k + 2));
					glTexCoord2s (MEMSR16 (addr + k + 0), MEMSR16 (addr + k + 2));
					k += 4;
					break;

				default:
					DEBUG (EVENT_EFATAL, "!!!! unemulated tex VAT %d", vat_tex0_format);
			}
		}

		// position must be the last one
		switch (vat_pos_format)
		{
			case VAT_FLOAT:
				DEBUG (EVENT_LOG, "....  pos (%.2f, %.2f, %.2f)", MEMRF (addr + l + 0), MEMRF (addr + l + 4), MEMRF (addr + l + 8));
				glVertex3f (MEMRF (addr + l + 0), MEMRF (addr + l + 4), MEMRF (addr + l + 8));
				break;

			case VAT_U16:
				if (vat_pos_count == 0)
				{
					DEBUG (EVENT_LOG, "....  pos (%d, %d)", MEMR16 (addr + l + 0), MEMR16 (addr + l + 2));
					glVertex2i (MEMR16 (addr + l + 0), MEMR16 (addr + l + 2));
				}
				else
				{
					DEBUG (EVENT_LOG, "....  pos (%d, %d, %d)", MEMR16 (addr + l + 0), MEMR16 (addr + l + 2), MEMR16 (addr + l + 4));
					glVertex3i (MEMR16 (addr + l + 0), MEMR16 (addr + l + 2), MEMR16 (addr + l + 4));
				}
				break;

			case VAT_S16:
				if (vat_pos_count == 0)
				{
					DEBUG (EVENT_LOG, "....  pos (%d, %d)", MEMSR16 (addr + l + 0), MEMSR16 (addr + l + 2));
					glVertex2s (MEMSR16 (addr + l + 0), MEMSR16 (addr + l + 2));
				}
				else
				{
					DEBUG (EVENT_LOG, "....  pos (%d, %d, %d)", MEMSR16 (addr + l + 0), MEMSR16 (addr + l + 2), MEMSR16 (addr + l + 4));
					glVertex3s (MEMSR16 (addr + l + 0), MEMSR16 (addr + l + 2), MEMSR16 (addr + l + 4));
				}
				break;

			default:
				DEBUG (EVENT_EFATAL, "!!!! unemulated position VAT %d", vat_pos_format);
		}
		l = k;
	}
	
	return k;
}


unsigned int gx_parse_command (__u32 addr)
{
	unsigned int i, n = 0, k = 0, a;


	switch (MEM (addr++))
	{
		// nop
		case 0x00:
			DEBUG (EVENT_LOG, "....  nop");
			return 4;

		// load cp
		case 0x08:
			{
				char *mode[] = { "OFF", "D", "I8", "I16" };
				__u32 data = MEMR32 (addr + 1);
			
				switch (MEM (addr))
				{
					// VCD_LO
					case 0x50:
						vcd_pos = (data >> 9) & 3;
						vcd_normal = (data >> 11) & 3;
						vcd_col0 = (data >> 13) & 3;
						vcd_col1 = (data >> 15) & 3;
						DEBUG (EVENT_LOG, "....  VCD pos %s nrm %s col0 %s col1 %s",
									 mode[vcd_pos], mode[vcd_normal], mode[vcd_col0], mode[vcd_col1]);
						break;
					
					// VCD_HI
					case 0x60:
						vcd_tex0 = data & 3;
						DEBUG (EVENT_LOG, "....  VCD tex0 %s", mode[vcd_tex0]);
						break;

					// VAT_A
					case 0x70:
						{
						__u32 data = MEMR32 (addr + 1);
							vat_pos_count = (data >> 0) & 1;
							vat_pos_format = (data >> 1) & 3;
							vat_normal_count = (data >> 9) & 1;
							vat_normal_format = (data >> 10) & 3;
							vat_col0_count = (data >> 13) & 1;
							vat_col0_format = (data >> 14) & 3;
							vat_tex0_count = (data >> 21) & 1;
							vat_tex0_format = (data >> 22) & 3;

							DEBUG (EVENT_LOG, "....  VAT pc %d pf %d nc %d nf %d c0c %d c0f %d t0c %d t0f %d",
										 vat_pos_count, vat_pos_format, vat_normal_count, vat_normal_format, vat_col0_count, vat_col0_format, vat_tex0_count, vat_tex0_format);
						}
						break;
					
					default:
						DEBUG (EVENT_LOG, "....  cp[%2.2x] = %8.8x", MEM (addr), data);
				}
			}
			return 6;

		// load xf
		case 0x10:
			{
				// length
				n = MEMR16 (addr) + 1;
				// 1st address
				a = MEMR16 (addr + 2);
				addr += 4;
				if (a == 0x1020)
				{
					float m[] = 
					{
						1, 0, 0, 0,				// A, 0, 0, 0,
						0, 1, 0, 0,				// 0, C, 0, 0,
						0, 0, 1, 0,				// 0, 0, D, 0,
						0, 0, 0, 1,				// B, D, F, 1,
					};
					int index[] = {0, 12, 5, 13, 10, 14};

					DEBUG (EVENT_LOG, "....  SET PROJECTION MATRIX (%d)", n);
					for (i = 0; i < n; i++, addr += 4)
					{
						DEBUG (EVENT_LOG, "....  PARAM_%c = %f", 'A' + (a + i) - 0x1020, MEMRF (addr));
						if (i < 6)
							m[index[i]] = MEMRF (addr);
					}
					glLoadMatrixf (m);
				}
				else
					for (i = 0; i < n; i++, addr += 4)
						DEBUG (EVENT_LOG, "....  xf[%4.4x] = %8.8x", a + i, MEMR32 (addr));
			}
			return (5 + n * 4);

		// call displaylist
		case 0x40:
			DEBUG (EVENT_LOG, "....  call displaylist: address %8.8x, size %8.8x",
						 MEMR32 (addr), MEMR32 (addr + 4));
			return 9;

		// invalidate vertex cache
		case 0x48:
			DEBUG (EVENT_LOG, "....  invalidate vertex cache");
			return 1;

		// load bp
		case 0x61:
			{
				__u32 data = MEMR32 (addr) & 0x00ffffff;

				switch (MEM (addr))
				{
					case 0x80:
						tex_mode = data;
						DEBUG (EVENT_LOG, "....  TEX0MODE_I0 | aniso %d | filter %d%d | wrap_s %d | wrap_t %d",
									 (tex_mode >> 19) & 3, (tex_mode & 0x20)>0, (tex_mode & 0x10)>0, (tex_mode >> 2) & 3, tex_mode & 3);
						break;

					case 0x88:
						tex_width = (data & 0x3ff) + 1;
						tex_height = ((data >> 10) & 0x3ff) + 1;
						tex_format = (data >> 20) & 0x7;
						DEBUG (EVENT_LOG, "....  TEX0_I0 | width %d | height %d | format %x",
									 tex_width, tex_height, tex_format);
						break;
					
					case 0x94:
						{
							unsigned int tex = 0;
							int i;
							__u32 mem = (data & 0x00ffffffff) << 5;


							for (i = 0; i < ntextures; i++)
								if (mem == mtextures[i])
								{
									tex = textures[i];
									break;
								}
							
							switch (tex_format)
							{
								case 0x04:	//
								case 0x06:	//

								case RGB5A1:
									if (!tex)
									{
										if (use_textures)
											glEnable (GL_TEXTURE_RECTANGLE_NV);
										glGenTextures (1, &tex);
										glBindTexture (GL_TEXTURE_RECTANGLE_NV, tex);

										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, TEX_MODE_MIN_FILTER (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, TEX_MODE_MAG_FILTER (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, TEX_MODE_WRAP_S (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, TEX_MODE_WRAP_T (tex_mode));

										glTexImage2D (GL_TEXTURE_RECTANGLE_NV, 0, GL_RGBA, tex_width, tex_height, 0, GL_RGBA, GL_UNSIGNED_BYTE,
																	gx_convert_texture (MEM_ADDRESS (mem), tex_width, tex_height, tex_format));
										textures[ntextures] = tex;
										mtextures[ntextures] = mem;
										ntextures++;
										DEBUG (EVENT_LOG, "....  TEX load from %.8x (%d)", mem | 0x80000000, tex);
									}
									else
									{
										if (use_textures)
											glEnable (GL_TEXTURE_RECTANGLE_NV);
										glBindTexture (GL_TEXTURE_RECTANGLE_NV, tex);
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, TEX_MODE_MIN_FILTER (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, TEX_MODE_MAG_FILTER (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, TEX_MODE_WRAP_S (tex_mode));
										glTexParameteri (GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, TEX_MODE_WRAP_T (tex_mode));
										DEBUG (EVENT_LOG, "....  TEX already loaded (%d)", tex);
									}
									break;
								
								default:
									DEBUG (EVENT_LOG, "!...  TEX unknown format (%.8x)", mem | 0x80000000);
							}
						}

					default:
						DEBUG (EVENT_LOG, "....  bp[%2.2x] = %6.6x", MEMR32 (addr) >> 24, MEMR32 (addr) & 0x00ffffff);
				}
			}
			return 5;

		// draw quads
		case 0x80:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw %d quads", n);

				glBegin (GL_QUADS);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw triangles
		case 0x90:
			return (3 + n * 16);
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw %d triangles", n);

				glBegin (GL_TRIANGLES);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw triangle strip
		case 0x98:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw triangle strip with %d vertices", n);

				glBegin (GL_TRIANGLE_STRIP);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw triangle fan
		case 0xa0:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw triangle fan with %d vertices", n);

				glBegin (GL_TRIANGLE_FAN);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw lines
		case 0xa8:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw %d lines", n);

				glBegin (GL_LINES);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw line strip
		case 0xb0:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw line strip with %d vertices", n);

				glBegin (GL_LINE_STRIP);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		// draw points
		case 0xb8:
			{
				n = MEMR16 (addr);
				addr += 2;

				DEBUG (EVENT_LOG, "....  draw %d points", n);

				glBegin (GL_POINTS);
				k = push_vertices (n, addr);
				glEnd ();
			}
			return (3 + k);

		default:
			DEBUG (EVENT_EFATAL, "..gx: unknown command %2.2x", MEM (addr-1));
			return 1;
	}
}


void gx_parse_fifo (__u32 fifo_start, __u32 fifo_end)
{
	DEBUG (EVENT_LOG, "..gx: parsing fifo (size = %.8x, base = %.8x, wpointer = %.8x)", fifo_end - fifo_start, CP_FIFO_BASE, CP_FIFO_WPOINTER);

	while (fifo_start < fifo_end)
	{
		fifo_start += gx_parse_command (fifo_start);
	}
}


void gx_w8_command (__u32 addr, __u8 data)
{
	if (!CP_FIFO_WPOINTER)
		return;

	MEMW (CP_FIFO_WPOINTER, data);
#ifdef LOG_GX_WRITE
	DEBUG (EVENT_LOG, "..gx: (%.8x) [%.8x] = %2.2x", PC, CP_FIFO_WPOINTER | 0x80000000, data);
#endif
	CP_FIFO_WPOINTER++;
}


void gx_w16_command (__u32 addr, __u16 data)
{
	if (!CP_FIFO_WPOINTER)
		return;

	MEMWR16 (CP_FIFO_WPOINTER, data);
#ifdef LOG_GX_WRITE
	DEBUG (EVENT_LOG, "..gx: (%.8x) [%.8x] = %4.4x", PC, CP_FIFO_WPOINTER | 0x80000000, data);
#endif
	CP_FIFO_WPOINTER += 2;
}


void gx_w32_command (__u32 addr, __u32 data)
{
	if (!CP_FIFO_WPOINTER)
		return;

	MEMWR32 (CP_FIFO_WPOINTER, data);
#ifdef LOG_GX_WRITE
	DEBUG (EVENT_LOG, "..gx: (%.8x) [%.8x] = %8.8x", PC, CP_FIFO_WPOINTER | 0x80000000, data);
#endif
	if ((MEMR16 (CP_FIFO_WPOINTER + 2) == 0x4700) &&
			(MEMR16 (CP_FIFO_WPOINTER - 3) == 0x4800))
		pe_token (MEMR16 (CP_FIFO_WPOINTER));
	else if ((data == 0x45000002) && (MEM (CP_FIFO_WPOINTER - 1) == 0x61))
	{
		pe_finish ();
		gx_parse_fifo (CP_FIFO_BASE | CP_FIFO_RW_DIST, CP_FIFO_WPOINTER + 4);
		CP_FIFO_WPOINTER = (CP_FIFO_BASE | CP_FIFO_RW_DIST) - 4;
	}

	CP_FIFO_WPOINTER += 4;
}


void gx_init (void)
{
	// gx command
	mem_hww_hook (8, 0x8000, gx_w8_command);
	mem_hww_hook (16, 0x8000, gx_w16_command);
	mem_hww_hook (32, 0x8000, gx_w32_command);

	// for paired singles store
	mem_hww_hook (8, 0x8004, gx_w8_command);
	mem_hww_hook (16, 0x8004, gx_w16_command);
	mem_hww_hook (32, 0x8004, gx_w32_command);
}
