/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  SI - serial interface (0xcc006400, 0x40)
 *      
 *         
 */

#include "hw_si.h"


__u8 rsi[RSI_SIZE] = {0};


__u32 chn_cmd[4] = { SI_CH0_CMD, SI_CH1_CMD, SI_CH2_CMD, SI_CH3_CMD };
__u32 chn_inl[4] = { SI_CH0_INL, SI_CH1_INL, SI_CH2_INL, SI_CH3_INL };
__u32 chn_inu[4] = { SI_CH0_INU, SI_CH1_INU, SI_CH2_INU, SI_CH3_INU };
#define SI_CHN_CMD(n)			((__u32) chn_cmd[n])
#define SI_CHN_INL(n)			((__u32) chn_inl[n])
#define SI_CHN_INU(n)			((__u32) chn_inu[n])



__u32 si_r32_direct (__u32 addr)
{
//	DEBUG (EVENT_LOG, "..si: read  [%.4x] (%.8x)", addr & 0xffff, RSI32 (addr));
	return RSI32 (addr);
}


void si_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..si: write [%.4x] (%.8x)", addr & 0xffff, data);
	RSI32 (addr) = data;
}


void si_w32_commcsr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..si: CSR = %.8x", data);

	// acknowlegde interrupt
	if (data & SI_COMMCSR_TCINT)
		RSI32 (addr) &= ~SI_COMMCSR_TCINT;

	if (data & SI_COMMCSR_TSTART)
	{
		// do transfer
		DEBUG (EVENT_LOG, "..si: COMMAND %.2x", RSI8 (SI_IO_BUFF));
		switch (RSI8 (SI_IO_BUFF))
		{
			// get device type and status
			case 0x00:
				// set first pad to default gc controller
				// the other pads are not important right now
				RSI8 (SI_IO_BUFF + 0) = 9;
				RSI8 (SI_IO_BUFF + 1) = 0;
				RSI8 (SI_IO_BUFF + 2) = 0;
				RSI8 (SI_IO_BUFF + 3) = 0;
				break;
			
			// read origins
			case 0x41:
				RSI8 (SI_IO_BUFF + 1) = 0;
				RSI8 (SI_IO_BUFF + 2) = 0x80;
				RSI8 (SI_IO_BUFF + 3) = 0x80;
				RSI8 (SI_IO_BUFF + 4) = 0x80;
				RSI8 (SI_IO_BUFF + 5) = 0x80;
				RSI8 (SI_IO_BUFF + 6) = 0x1f;
				RSI8 (SI_IO_BUFF + 7) = 0x1f;
				break;
		}
		
		// complete transfer
		RSI32 (addr) &= ~SI_COMMCSR_TSTART;

		// transfer complete interrupt
		RSI32 (addr) |= SI_COMMCSR_TCINT;
		if (data & SI_COMMCSR_TCINTMSK)
		{
			RSI32 (addr) |= SI_COMMCSR_TCINTMSK;
			pi_interrupt (INTERRUPT_SI);
		}
	}
}


void si_w32_sr (__u32 addr, __u32 data)
{
	data &= ~(SI_SR_WRST0 | SI_SR_WRST1 | SI_SR_WRST2 | SI_SR_WRST3);

	DEBUG (EVENT_LOG, "..si: SR = %.8x", data);
	RSI32 (addr) = data;
}


void pad_set_action (int n, __u16 buttons, __s8 xstick, __s8 ystick,
										 __s8 cxstick, __s8 cystick, __s8 trigl, __s8 trigr)
{
	RSI32 (SI_CHN_INL (n)) = SI_CHL (buttons, xstick, ystick);
	RSI32 (SI_CHN_INU (n)) = SI_CHU (cxstick, cystick, trigl, trigr);
}


// for keyboard mapping
void pad_action (int n, int pressed, int action)
{
	if (action < ACTION_AXIS)
	{
		pad_action_buttons (n, pressed, action);
		return;
	}

	action -= ACTION_AXIS;
	// axis0-, axis0+, ...
	if (pressed)
		pad_action_axis (n, action / 2, (action % 2 == 1) ? 127 : -127);
	else
		pad_action_axis (n, action / 2, 0);
}


void pad_action_buttons (int n, int pressed, __u16 buttons)
{
	if (pressed)
		RSI32 (SI_CHN_INL (n)) |=  (buttons << 16);
	else
		RSI32 (SI_CHN_INL (n)) &= ~(buttons << 16);
}


void pad_action_buttons_masked (int n, int pressed, __u16 buttons, int mask)
{
	RSI32 (SI_CHN_INL (n)) &= ~(mask << 16);
	pad_action_buttons (n, pressed, buttons);
}


void pad_action_analogs (int n, __s8 x, __s8 y)
{
	RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0xffff0000) |
													((__u8) x << 8) | ((__u8) y);
}


void pad_action_canalogs (int n, __s8 x, __s8 y)
{
	RSI32 (SI_CHN_INU (n)) = (RSI32 (SI_CHN_INU (n)) & 0x0000ffff) |
													((__u8) x << 24) | ((__u8) y << 16);
}


void pad_action_triggers (int n, __s8 l, __s8 r)
{
	RSI32 (SI_CHN_INU (n)) = (RSI32 (SI_CHN_INU (n)) & 0xffff0000) |
													((__u8) l << 8) | ((__u8) r);
}


void pad_action_axis (int n, int axis, __s8 value)
{
	switch (axis)
	{
		// analog y
		case 0:
			RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0xffffff00) |
															((__u8) (0x80 - value));
			break;

		// analog x
		case 1:
			RSI32 (SI_CHN_INL (n)) = (RSI32 (SI_CHN_INL (n)) & 0xffff00ff) |
															((__u8) (0x80 + value) << 8);
			break;

		// canalog y
		case 2:
			RSI32 (SI_CHN_INU (n)) = (RSI32 (SI_CHN_INU (n)) & 0xff00ffff) |
															((__u8) (0x80 - value) << 16);
			break;
		
		// canalog x
		case 3:
			RSI32 (SI_CHN_INU (n)) = (RSI32 (SI_CHN_INU (n)) & 0x00ffffff) |
															((__u8) (0x80 + value) << 24);
			break;

		default:
			break;
	}
}


void si_poll (void)
{
	if (RSI32 (SI_POLL) & SI_POLL_EN0)
	{
		RSI32 (SI_SR) |= SI_SR_RDST0;
		RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN1)
	{
		RSI32 (SI_SR) |= SI_SR_RDST1;
		RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN2)
	{
		RSI32 (SI_SR) |= SI_SR_RDST2;
		RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
	}

	if (RSI32 (SI_POLL) & SI_POLL_EN3)
	{
		RSI32 (SI_SR) |= SI_SR_RDST3;
		RSI32 (SI_COMMCSR) |= SI_COMMCSR_RDSTINT;
	}
}


void si_init (void)
{
	int i;


	memset (rsi, 0, sizeof (rsi));

	// initialize pads
	for (i = 0; i < 4; i++)
	{
		// pad command
		RSI32 (SI_CHN_CMD (i)) = 0x00400300;		// ?

		// center analogs
		RSI32 (SI_CHN_INL (i)) = SI_CHL (0, 0, 0);
		RSI32 (SI_CHN_INU (i)) = SI_CHU (0, 0, 0, 0);
		mem_hwr_hook (32, SI_CHN_CMD (i), si_r32_direct);
		mem_hww_hook (32, SI_CHN_CMD (i), si_w32_direct);
		mem_hwr_hook (32, SI_CHN_INL (i), si_r32_direct);
		mem_hww_hook (32, SI_CHN_INL (i), si_w32_direct);
		mem_hwr_hook (32, SI_CHN_INU (i), si_r32_direct);
		mem_hww_hook (32, SI_CHN_INU (i), si_w32_direct);
	}

	// sipoll
	mem_hwr_hook (32, SI_POLL, si_r32_direct);	
	mem_hww_hook (32, SI_POLL, si_w32_direct);	

	// sicommcsr
	mem_hwr_hook (32, SI_COMMCSR, si_r32_direct);	
	mem_hww_hook (32, SI_COMMCSR, si_w32_commcsr);	

	// sisr - si status register
	mem_hwr_hook (32, SI_SR, si_r32_direct);	
	mem_hww_hook (32, SI_SR, si_w32_sr);	

	// si io buffer
	for (i = 0; i < 128; i += 4)
	{
		mem_hww_hook (32, SI_IO_BUFF + i, si_w32_direct);
		mem_hwr_hook (32, SI_IO_BUFF + i, si_r32_direct);
	}
}
